#pragma once

#include "BaseQueryProcessorBase.h"

#include <RobotAPI/libraries/armem/core/Time.h>
#include <RobotAPI/libraries/armem/core/error.h>
#include <RobotAPI/libraries/armem/core/ice_conversions.h>

#include <RobotAPI/interface/armem/query.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/logging/Logging.h>

#include <cstdint>
#include <iterator>


namespace armarx::armem::server::query_proc::base
{

    template <QueryTarget queryTarget, class _EntityT, class _ResultEntityT>
    class EntityQueryProcessorBase :
        public BaseQueryProcessorBase<queryTarget, _EntityT, _ResultEntityT, armem::query::data::EntityQuery>
    {
        using Base = BaseQueryProcessorBase<queryTarget, _EntityT, _ResultEntityT, armem::query::data::EntityQuery>;

    public:

        using EntityT = _EntityT;
        using EntitySnapshotT = typename EntityT::EntitySnapshotT;

        using ResultEntityT = _ResultEntityT;
        using ResultSnapshotT = typename ResultEntityT::EntitySnapshotT;


        using Base::process;
        virtual void process(ResultEntityT& result,
                             const armem::query::data::EntityQuery& query,
                             const EntityT& entity) const override
        {
            if (auto q = dynamic_cast<const armem::query::data::entity::All*>(&query))
            {
                process(result, *q, entity);
            }
            else if (auto q = dynamic_cast<const armem::query::data::entity::Single*>(&query))
            {
                process(result, *q, entity);
            }
            else if (auto q = dynamic_cast<const armem::query::data::entity::TimeRange*>(&query))
            {
                process(result, *q, entity);
            }
            else if (auto q = dynamic_cast<const armem::query::data::entity::IndexRange*>(&query))
            {
                process(result, *q, entity);
            }
            else if (auto q = dynamic_cast<const armem::query::data::entity::TimeApprox*>(&query))
            {
                process(result, *q, entity);
            }
            else if (auto q = dynamic_cast<const armem::query::data::entity::BeforeOrAtTime*>(&query))
            {
                process(result, *q, entity);
            }
            else if (auto q = dynamic_cast<const armem::query::data::entity::BeforeTime*>(&query))
            {
                process(result, *q, entity);
            }
            else
            {
                throw armem::error::UnknownQueryType("entity snapshot", query);
            }
        }


        void process(ResultEntityT& result,
                     const armem::query::data::entity::All& query,
                     const EntityT& entity) const
        {
            (void) query;
            // Copy this entitiy and its contents.

            entity.forEachSnapshot([this, &result](const EntitySnapshotT & snapshot)
            {
                addResultSnapshot(result, snapshot);
            });
        }


        void process(ResultEntityT& result,
                     const armem::query::data::entity::Single& query,
                     const EntityT& entity) const
        {
            if (query.timestamp < 0)
            {
                try
                {
                    addResultSnapshot(result, entity.getLatestSnapshot());
                }
                catch (const armem::error::EntityHistoryEmpty&)
                {
                    if (false)
                    {
                        ARMARX_IMPORTANT << "Failed to retrieve latest snapshot from entity " << entity.id() << ". "
                                         << "Entity has not timestamps.";
                    }
                }
            }
            else
            {
                const Time time = fromIce<Time>(query.timestamp);
                try
                {
                    auto snapshot = entity.getSnapshot(time);
                    addResultSnapshot(result, snapshot);
                }
                catch (const armem::error::MissingEntry&)
                {
                    // Leave empty.
                    if (false)
                    {
                        std::stringstream ss;
                        ss << "Failed to retrieve snapshot with timestamp "
                           << armem::toDateTimeMilliSeconds(time)
                           << " from entity " << entity.id() << ".\n"
                           << "Entity has timestamps: ";
                        for (const Time& t : entity.getTimestamps())
                        {
                            ss << "\n- " << armem::toDateTimeMilliSeconds(t);
                        }
                        ARMARX_IMPORTANT << ss.str();
                    }
                }
            }
        }


        void process(ResultEntityT& result,
                     const armem::query::data::entity::TimeRange& query,
                     const EntityT& entity) const
        {
            if (query.minTimestamp <= query.maxTimestamp || query.minTimestamp < 0 || query.maxTimestamp < 0)
            {
                Time min = fromIce<Time>(query.minTimestamp);
                Time max = fromIce<Time>(query.maxTimestamp);
                process(result, min, max, entity, query);
            }
        }


        void process(ResultEntityT& result,
                     const armem::query::data::entity::IndexRange& query,
                     const EntityT& entity) const
        {
            entity.forEachSnapshotInIndexRange(
                query.first, query.last,
                [this, &result](const EntitySnapshotT & snapshot)
            {
                addResultSnapshot(result, snapshot);
            });
        }


        void process(ResultEntityT& result,
                     const Time& min,
                     const Time& max,
                     const EntityT& entity,
                     const armem::query::data::EntityQuery& query) const
        {
            (void) query;
            entity.forEachSnapshotInTimeRange(
                min, max,
                [this, &result](const EntitySnapshotT & snapshot)
            {
                addResultSnapshot(result, snapshot);
            });
        }


        void process(ResultEntityT& result,
                     const armem::query::data::entity::BeforeOrAtTime& query,
                     const EntityT& entity) const
        {
            const auto referenceTimestamp = fromIce<Time>(query.timestamp);
            ARMARX_CHECK(referenceTimestamp.toMicroSeconds() >= 0) << "Reference timestamp is negative!";

            if (auto* beforeOrAt = entity.findLatestSnapshotBeforeOrAt(referenceTimestamp))
            {
                addResultSnapshot(result, *beforeOrAt);
            }
        }


        void process(ResultEntityT& result,
                     const armem::query::data::entity::BeforeTime& query,
                     const EntityT& entity) const
        {
            const armem::Time referenceTimestamp = fromIce<Time>(query.timestamp);
            ARMARX_CHECK(referenceTimestamp.toMicroSeconds() >= 0) << "Reference timestamp must be non-negative.";

            std::vector<const EntitySnapshotT*> befores;
            entity.forEachSnapshotBefore(referenceTimestamp, [&befores](const EntitySnapshotT & s)
            {
                befores.push_back(&s);
            });

            size_t num = 0;
            if (query.maxEntries < 0)
            {
                num = befores.size();
            }
            else
            {
                num = std::min(befores.size(), static_cast<size_t>(query.maxEntries));
            }

            for (size_t r = 0; r < num; ++r)
            {
                size_t i = befores.size() - 1 - r;
                addResultSnapshot(result, *befores[i]);
            }
        }


        void process(ResultEntityT& result,
                     const armem::query::data::entity::TimeApprox& query,
                     const EntityT& entity) const
        {
            const Time referenceTimestamp = fromIce<Time>(query.timestamp);
            ARMARX_CHECK(referenceTimestamp.toMicroSeconds() >= 0) << "Reference timestamp is negative!";

            const float referenceTimestampMicroSeconds = referenceTimestamp.toMicroSeconds();
            const float epsDuration = fromIce<Time>(query.eps).toMicroSeconds();

            // elements have to be in range [t_ref - eps, t_ref + eps] if eps is positive
            const auto isInRange = [&](const Time & t) -> bool
            {
                if (epsDuration <= 0)
                {
                    return true;
                }

                return std::abs(t.toMicroSeconds() - referenceTimestampMicroSeconds) <= epsDuration;
            };

            // last element before or at timestamp
            if (auto* beforeOrAt = entity.findLatestSnapshotBeforeOrAt(referenceTimestamp))
            {
                const auto timestampOfMatchBefore = beforeOrAt->id().timestamp;
                const auto isPerfectMatch = timestampOfMatchBefore == referenceTimestamp;
                if (isInRange(timestampOfMatchBefore))
                {
                    addResultSnapshot(result, *beforeOrAt);
                }

                // earsly stop, not necessary to also get the next since the match is perfect
                if (isPerfectMatch)
                {
                    return;
                }

                // first element after or at timestamp (or at because of fewer checks, we can assure that there is not element at)
                const auto* after = entity.findFirstSnapshotAfterOrAt(referenceTimestamp);
                if (after)
                {
                    const auto timestampOfMatchAfter = after->id().timestamp;
                    if (isInRange(timestampOfMatchAfter))
                    {
                        addResultSnapshot(result, *after);
                    }
                }
            }
        }


    protected:

        void addResultSnapshot(ResultEntityT& result, const EntitySnapshotT& snapshot) const
        {
            result.addSnapshot(snapshot);
        }

    };
}
