#pragma once

#include "BaseQueryProcessorBase.h"

#include <RobotAPI/libraries/armem/core/error.h>
#include <RobotAPI/interface/armem/query.h>

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <regex>


namespace armarx::armem::server::query_proc::base
{

    template <QueryTarget queryTarget, class _MemoryT, class _ResultMemoryT, class _ChildProcessorT>
    class MemoryQueryProcessorBase :
        public BaseQueryProcessorBase<queryTarget, _MemoryT, _ResultMemoryT, armem::query::data::MemoryQuery>
    {
        using Base = BaseQueryProcessorBase<queryTarget, _MemoryT, _ResultMemoryT, armem::query::data::MemoryQuery>;

    public:

        using MemoryT = _MemoryT;
        using CoreSegmentT = typename MemoryT::CoreSegmentT;

        using ResultMemoryT = _ResultMemoryT;
        using ResultCoreSegmentT = typename ResultMemoryT::CoreSegmentT;

        using ChildProcessorT = _ChildProcessorT;


    public:

        MemoryQueryProcessorBase()
        {
        }
        MemoryQueryProcessorBase(ChildProcessorT&& childProcessor) :
            childProcessor(childProcessor)
        {
        }


        using Base::process;
        ResultMemoryT process(const armem::query::data::Input& input, const MemoryT& memory) const
        {
            return this->process(input.memoryQueries, memory);
        }

        virtual void process(ResultMemoryT& result,
                             const armem::query::data::MemoryQuery& query,
                             const MemoryT& memory) const override
        {
            if (auto q = dynamic_cast<const armem::query::data::memory::All*>(&query))
            {
                process(result, *q, memory);
            }
            else if (auto q = dynamic_cast<const armem::query::data::memory::Single*>(&query))
            {
                process(result, *q, memory);
            }
            else if (auto q = dynamic_cast<const armem::query::data::memory::Regex*>(&query))
            {
                process(result, *q, memory);
            }
            else
            {
                throw armem::error::UnknownQueryType(MemoryT::getLevelName(), query);
            }
        }

        void process(ResultMemoryT& result,
                     const armem::query::data::memory::All& query,
                     const MemoryT& memory) const
        {
            memory.forEachCoreSegment([this, &result, &query](const CoreSegmentT & coreSegment)
            {
                childProcessor.process(result.addCoreSegment(coreSegment.name()), query.coreSegmentQueries, coreSegment);
            });
        }

        void process(ResultMemoryT& result,
                     const armem::query::data::memory::Single& query,
                     const MemoryT& memory) const
        {
            try
            {
                const CoreSegmentT& coreSegment = memory.getCoreSegment(query.coreSegmentName);
                childProcessor.process(result.addCoreSegment(coreSegment.name()), query.coreSegmentQueries, coreSegment);
            }
            catch (const error::MissingEntry&)
            {
            }
        }

        void process(ResultMemoryT& result,
                     const armem::query::data::memory::Regex& query,
                     const MemoryT& memory) const
        {
            const std::regex regex(query.coreSegmentNameRegex);
            memory.forEachCoreSegment([this, &result, &query, &regex](const CoreSegmentT & coreSegment)
            {
                if (std::regex_search(coreSegment.name(), regex))
                {
                    childProcessor.process(result.addCoreSegment(coreSegment.name()), query.coreSegmentQueries, coreSegment);
                }
            });
        }


    protected:

        ChildProcessorT childProcessor;

    };
}
