#include "InstanceView.h"

#include <QAction>
#include <QApplication>
#include <QClipboard>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QHeaderView>
#include <QImage>
#include <QLabel>
#include <QLayout>
#include <QMenu>
#include <QSplitter>
#include <QTreeWidget>
#include <QVBoxLayout>

#include <SimoxUtility/algorithm/string.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/libraries/aron/core/navigator/data/complex/NDArray.h>
#include <RobotAPI/libraries/aron/core/navigator/type/container/Object.h>

#include <RobotAPI/libraries/armem/aron/MemoryID.aron.generated.h>
#include <RobotAPI/libraries/armem/core/aron_conversions.h>

#include <RobotAPI/libraries/armem_gui/gui_utils.h>
#include <RobotAPI/libraries/armem_gui/instance/ImageView.h>
#include <RobotAPI/libraries/armem_gui/instance/sanitize_typename.h>
#include <RobotAPI/libraries/armem_gui/instance/serialize_path.h>
#include <RobotAPI/libraries/armem_gui/instance/tree_builders/DataTreeBuilder.h>
#include <RobotAPI/libraries/armem_gui/instance/tree_builders/TypedDataTreeBuilder.h>

#include "MemoryIDTreeWidgetItem.h"
#include "WidgetsWithToolbar.h"


namespace armarx::armem::gui::instance
{

    InstanceView::InstanceView()
    {
        Logging::setTag("InstanceView");

        QLayout* layout = new QVBoxLayout();
        this->setLayout(layout);
        int margin = 3;
        layout->setContentsMargins(margin, margin, margin, margin);

        splitter = new QSplitter(Qt::Orientation::Vertical);
        layout->addWidget(splitter);

        tree = new QTreeWidget(this);
        splitter->addWidget(tree);

        QStringList columns;
        columns.insert(int(Columns::KEY), "Key");
        columns.insert(int(Columns::VALUE), "Value");
        columns.insert(int(Columns::TYPE), "Type");
        tree->setColumnCount(columns.size());
        tree->setHeaderLabels(columns);

        tree->header()->resizeSection(int(Columns::KEY), 250);
        tree->header()->resizeSection(int(Columns::VALUE), 200);

        treeItemInstanceID = new MemoryIDTreeWidgetItem({"Instance ID"});
        treeItemInstanceID->addKeyChildren();

        treeItemMetadata = new QTreeWidgetItem({"Metadata"});
        treeItemMetadata->addChild(new QTreeWidgetItem({"Confidence"}));
        treeItemMetadata->addChild(new QTreeWidgetItem({"Time Created"}));
        treeItemMetadata->addChild(new QTreeWidgetItem({"Time Sent"}));
        treeItemMetadata->addChild(new QTreeWidgetItem({"Time Arrived"}));

        treeItemData = new QTreeWidgetItem({"Data"});

        QList<QTreeWidgetItem*> items = {treeItemInstanceID, treeItemMetadata, treeItemData};
        tree->addTopLevelItems(items);
        for (auto* item : items)
        {
            item->setExpanded(true);
        }
        treeItemMetadata->setExpanded(false);

        tree->setContextMenuPolicy(Qt::CustomContextMenu);
        connect(tree, &QTreeWidget::customContextMenuRequested, this, &This::prepareTreeContextMenu);
    }

    void InstanceView::setStatusLabel(QLabel* statusLabel)
    {
        this->statusLabel = statusLabel;
    }

    void InstanceView::setUseTypeInfo(bool enable)
    {
        this->useTypeInfo = enable;
        update();
    }


    void InstanceView::update(const MemoryID& id, const wm::Memory& memory)
    {
        aron::typenavigator::ObjectNavigatorPtr aronType = nullptr;
        const armem::wm::EntityInstance* instance = nullptr;
        try
        {
            instance = &memory.getInstance(id);
            if (useTypeInfo)
            {
                aronType = memory.getProviderSegment(id).aronType();
            }
        }
        catch (const armem::error::ArMemError& e)
        {
            showErrorMessage(e.what());
            return;
        };
        if (instance)
        {
            update(*instance, aronType);
        }
    }


    void InstanceView::update(const wm::EntityInstance& instance, aron::typenavigator::ObjectNavigatorPtr aronType)
    {
        currentInstance = instance;
        currentAronType = aronType;
        update();
    }


    void InstanceView::update()
    {
        if (currentInstance)
        {
            updateInstanceID(currentInstance->id());
            updateMetaData(currentInstance->metadata());
            updateData(currentInstance->data(), currentAronType);
            updateImageView(currentInstance->data());

            emit updated();
        }
    }


    void InstanceView::addInstanceView(const wm::EntityInstance& instance, aron::typenavigator::ObjectNavigatorPtr aronType)
    {
        ARMARX_IMPORTANT << "Adding instance view with toolbar for instance: " << instance.id();
        InstanceView* view = new InstanceView;
        view->setStatusLabel(statusLabel);
        view->setUseTypeInfo(useTypeInfo);

        WidgetsWithToolbar* child = new WidgetsWithToolbar();
        child->addWidget(view);


        splitter->addWidget(child);

        // Propagate this signal upwards.
        connect(view, &InstanceView::memoryIdResolutionRequested, this, &This::memoryIdResolutionRequested);

        view->update(instance, aronType);
    }


    void InstanceView::updateInstanceID(const MemoryID& id)
    {
        treeItemInstanceID->setInstanceID(id, int(Columns::VALUE));
    }


    void InstanceView::updateData(
        const aron::datanavigator::DictNavigatorPtr& data, aron::typenavigator::ObjectNavigatorPtr aronType)
    {
        if (!data)
        {
            armarx::gui::clearItem(treeItemData);
            QTreeWidgetItem* item = new QTreeWidgetItem({"(No data.)"});
            treeItemData->addChild(item);
        }
        else if (useTypeInfo && aronType)
        {
            treeItemData->setText(int(Columns::TYPE), QString::fromStdString(sanitizeTypeName(aronType->getName())));

            TypedDataTreeBuilder builder;
            builder.setColumns(int(Columns::KEY), int(Columns::VALUE), int(Columns::TYPE));
            builder.updateTree(treeItemData, *aronType, *data);
        }
        else
        {
            DataTreeBuilder builder;
            builder.setColumns(int(Columns::KEY), int(Columns::VALUE), int(Columns::TYPE));
            builder.updateTree(treeItemData, data);
        }
        treeItemData->setExpanded(true);
    }

    void InstanceView::updateMetaData(const wm::EntityInstanceMetadata& metadata)
    {
        std::vector<std::string> items =
        {
            std::to_string(metadata.confidence),
            armem::toDateTimeMilliSeconds(metadata.timeCreated),
            armem::toDateTimeMilliSeconds(metadata.timeSent),
            armem::toDateTimeMilliSeconds(metadata.timeArrived)
        };
        ARMARX_CHECK_EQUAL(treeItemMetadata->childCount(), items.size());
        int i = 0;
        for (const std::string& item : items)
        {
            treeItemMetadata->child(i++)->setText(int(Columns::VALUE), QString::fromStdString(item));
        }
    }

    void InstanceView::showErrorMessage(const std::string& message)
    {
        if (statusLabel)
        {
            statusLabel->setText(QString::fromStdString(message));
        }
    }


    aron::Path InstanceView::getElementPath(const QTreeWidgetItem* item) const
    {
        QStringList qpath = item->data(int(Columns::KEY), Qt::UserRole).toStringList();
        aron::Path path = deserializePath(qpath);
        return path;
    }


    void InstanceView::prepareTreeContextMenu(const QPoint& pos)
    {
        const QTreeWidgetItem* item = tree->itemAt(pos);
        if (item == nullptr)
        {
            return;  // No item => no context menu.
        }

        QMenu menu(this);

        if (item == this->treeItemInstanceID && currentInstance.has_value())
        {
            if (QAction* action = makeActionCopyMemoryID(currentInstance->id()))
            {
                menu.addAction(action);
            }
        }
        else if (item->parent() == nullptr)
        {
            return;  // Other top level item => no context menu.
        }

        // Type descriptor based actions
        aron::type::Descriptor type = static_cast<aron::type::Descriptor>(item->data(int(Columns::TYPE), Qt::UserRole).toInt());
        switch (type)
        {
            case aron::type::Descriptor::eIVTCByteImage:
            {
                const aron::Path path = getElementPath(item);

                QAction* viewAction = new QAction("Show image");
                menu.addAction(viewAction);
                connect(viewAction, &QAction::triggered, [this, path]()
                {
                    this->showImageView(path);
                });
            }
            break;
            default:
                break;
        }

        // Type name based actions
        const std::string typeName = item->text(int(Columns::TYPE)).toStdString();
        if (typeName == instance::sanitizedMemoryIDTypeName)
        {
            const aron::Path path = getElementPath(item);

            if (std::optional<MemoryID> id = getElementMemoryID(path))
            {
                if (QAction* action = makeActionCopyMemoryID(id.value()))
                {
                    menu.addAction(action);
                }
                if (QAction* action = makeActionResolveMemoryID(id.value()))
                {
                    menu.addAction(action);
                }
            }
        }


        if (menu.actions().size() > 0)
        {
            menu.exec(tree->mapToGlobal(pos));
        }
    }


    std::optional<MemoryID> InstanceView::getElementMemoryID(const aron::Path& elementPath)
    {
        aron::datanavigator::NavigatorPtr element;
        try
        {
            element = currentInstance->data()->navigateAbsolute(elementPath);
        }
        // This can happen when the underlying entity structure changes (a new entity has been selected).
        catch (const aron::error::AronException&)
        {
            // showErrorMessage(e.what());
            return std::nullopt;
        }
        catch (const armarx::LocalException& e)
        {
            showErrorMessage(e.what());
            return std::nullopt;
        }

        std::stringstream couldNotParseMsg;
        couldNotParseMsg << "Element " << elementPath.toString() << " could not be parsed as MemoryID.";

        auto dictElement = std::dynamic_pointer_cast<aron::datanavigator::DictNavigator>(element);
        if (!dictElement)
        {
            showErrorMessage(couldNotParseMsg.str() + " (Failed to cast to DictNavigator.)");
            return std::nullopt;
        }

        try
        {
            arondto::MemoryID dto;
            dto.fromAron(dictElement);

            MemoryID id;
            armem::fromAron(dto, id);
            return id;
        }
        catch (const armarx::aron::error::AronException&)
        {
            showErrorMessage(couldNotParseMsg.str());
            return std::nullopt;
        }
    }


    QAction* InstanceView::makeActionResolveMemoryID(const MemoryID& id)
    {
        QAction* action = new QAction("Resolve memory ID");

        if (not(id.hasEntityName() and id.isWellDefined()))
        {
            action->setDisabled(true);
            action->setText(action->text() + " (incomplete Memory ID)");
        }
        connect(action, &QAction::triggered, [this, id]()
        {
            // ARMARX_IMPORTANT << "emit memoryIdResolutionRequested(id = " << id << ")";
            emit memoryIdResolutionRequested(id);
        });

        return action;
    }

    QAction* InstanceView::makeActionCopyMemoryID(const MemoryID& id)
    {
        QAction* action = new QAction("Copy memory ID to clipboard");

        connect(action, &QAction::triggered, [/*this,*/ id]()  // `this` for ARMARX_IMPORTANT
        {
            const QString idStr = QString::fromStdString(id.str());

            // ARMARX_IMPORTANT << "Copy '" << idStr.toStdString() << "' to clipboard.";
            QClipboard* clipboard = QApplication::clipboard();
            clipboard->setText(idStr);
            QApplication::processEvents();
        });

        return action;
    }


    void InstanceView::showImageView(const aron::Path& elementPath)
    {
        if (!currentInstance)
        {
            return;
        }
        if (!imageView)
        {
            WidgetsWithToolbar* toolbar = new WidgetsWithToolbar();

            imageView = new ImageView();
            imageView->toolbar = toolbar;
            toolbar->addWidget(imageView);

            splitter->addWidget(toolbar);

            connect(toolbar, &WidgetsWithToolbar::closing, [this]()
            {
                imageView = nullptr;
            });
        }
        imageView->elementPath = elementPath;
        updateImageView(currentInstance->data());
    }

    void InstanceView::removeImageView()
    {
        imageView->toolbar->close();
        imageView = nullptr;
    }


    void InstanceView::updateImageView(const aron::datanavigator::DictNavigatorPtr& data)
    {
        using aron::datanavigator::NDArrayNavigator;

        if (!imageView)
        {
            return;
        }
        if (!data)
        {
            removeImageView();
            return;
        }

        aron::datanavigator::NavigatorPtr element;
        try
        {
            element = data->navigateAbsolute(imageView->elementPath);
        }
        // This can happen when the underlying entity structure changes (a new entity has been selected).
        // In this case, we disable the image view.
        catch (const aron::error::AronException&)
        {
            // showErrorMessage(e.what());
            removeImageView();
            return;
        }
        catch (const armarx::LocalException&)
        {
            // showErrorMessage(e.what());
            removeImageView();
            return;
        }

        NDArrayNavigator::PointerType imageData = NDArrayNavigator::DynamicCast(element);
        if (!imageData)
        {
            showErrorMessage("Expected NDArrayNavigator, but got: " + simox::meta::get_type_name(element));
            return;
        }

        std::vector<int> shape = imageData->getDimensions();
        if (shape.size() != 3)
        {
            showErrorMessage("Expected array shape with 3 dimensions, but got: "
                             + NDArrayNavigator::DimensionsToString(shape));
            return;
        }

        QImage::Format format = QImage::Format::Format_Invalid;
        switch (shape.at(2))
        {
            case 1:
                format = QImage::Format::Format_Grayscale8;
                break;
            case 3:
                format = QImage::Format::Format_RGB888;
                break;
            default:
                showErrorMessage("Expected 1 or 3 elements in last dimension, but got shape: "
                                 + NDArrayNavigator::DimensionsToString(shape));
                return;
        }

        std::stringstream title;
        title << "Image element '" << imageView->elementPath.toString() << "'"; // of entity instance " << currentInstance->id();
        imageView->setTitle(QString::fromStdString(title.str()));
        imageView->view->setImage(QImage(imageData->getData(), shape.at(0), shape.at(1), format));
    }


    InstanceView::ImageView::ImageView()
    {
        setLayout(new QHBoxLayout());
        int margin = 2;
        layout()->setContentsMargins(margin, margin, margin, margin);
        if (false)
        {
            QFont font = this->font();
            font.setPointSizeF(font.pointSize() * 0.75);
            setFont(font);
        }

        view = new instance::ImageView();
        layout()->addWidget(view);
    }

}
