#pragma once

#include <sstream>
#include <stack>

#include <QTreeWidget>
#include <QLabel>

#include <RobotAPI/libraries/aron/core/navigator/visitors/TypedDataVisitor.h>

#include <RobotAPI/libraries/armem_gui/instance/sanitize_typename.h>
#include <RobotAPI/libraries/armem_gui/instance/serialize_path.h>

#include <RobotAPI/libraries/armem_gui/instance/tree_visitors/TreeDataVisitorBase.h>

#include <SimoxUtility/algorithm/string/string_tools.h>


namespace armarx::armem::gui::instance
{

    class TreeTypedDataVisitor :
        public aron::visitor::TypedDataVisitor,
        public TreeDataVisitorBase
    {
    public:

        using TreeDataVisitorBase::TreeDataVisitorBase;


        bool visitEnter(DictTypeNavigator& type, const std::string& key, DictDataNavigator& data) override
        {
            return _visitEnter(key, sanitizeTypeName(type.getName()), data.childrenSize());
        }
        bool visitExit(DictTypeNavigator&, DictDataNavigator&) override
        {
            return _visitExit();
        }

        bool visitEnter(ObjectTypeNavigator& type, const std::string& key, DictDataNavigator& data) override
        {
            return _visitEnter(key, sanitizeTypeName(type.getName()), data.childrenSize());
        }
        bool visitExit(ObjectTypeNavigator&, DictDataNavigator&) override
        {
            return _visitExit();
        }

        bool visitEnter(ListTypeNavigator& type, const std::string& key, ListDataNavigator& data) override
        {
            return _visitEnter(key, sanitizeTypeName(type.getName()), data.childrenSize());
        }
        bool visitExit(ListTypeNavigator&, ListDataNavigator&) override
        {
            return _visitExit();
        }

        bool visitEnter(TupleTypeNavigator& type, const std::string& key, ListDataNavigator& data) override
        {
            return _visitEnter(key, sanitizeTypeName(type.getName()), data.childrenSize());
        }
        bool visitExit(TupleTypeNavigator&, ListDataNavigator&) override
        {
            return _visitExit();
        }


        bool visit(BoolTypeNavigator& type, const std::string& key, BoolDataNavigator& data) override
        {
            return this->addValueRow(key, data, type);
        }
        bool visit(DoubleTypeNavigator& type, const std::string& key, DoubleDataNavigator& data) override
        {
            return this->addValueRow(key, data, type);
        }
        bool visit(FloatTypeNavigator& type, const std::string& key, FloatDataNavigator& data) override
        {
            return this->addValueRow(key, data, type);
        }
        bool visit(IntTypeNavigator& type, const std::string& key, IntDataNavigator& data) override
        {
            return this->addValueRow(key, data, type);
        }
        bool visit(LongTypeNavigator& type, const std::string& key, LongDataNavigator& data) override
        {
            return this->addValueRow(key, data, type);
        }
        bool visit(StringTypeNavigator& type, const std::string& key, StringDataNavigator& data) override
        {
            return this->addValueRow(key, data, type);
        }
        bool visit(TimeTypeNavigator& type, const std::string& key, LongDataNavigator& data) override
        {
            return this->addValueRow(key, data, type);
        }


        bool visit(EigenMatrixTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data) override
        {
            return this->addValueRow(key, data, type);
        }
        bool visit(EigenQuaternionTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data) override
        {
            return this->addValueRow(key, data, type);
        }
        bool visit(IVTCByteImageTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data) override
        {
            return this->addValueRow(key, data, type);
        }
        bool visit(OpenCVMatTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data) override
        {
            return this->addValueRow(key, data, type);
        }
        bool visit(PCLPointCloudTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data) override
        {
            return this->addValueRow(key, data, type);
        }
        bool visit(PoseTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data) override
        {
            return this->addValueRow(key, data, type);
        }
        bool visit(PositionTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data) override
        {
            return this->addValueRow(key, data, type);
        }
        bool visit(OrientationTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data) override
        {
            return this->addValueRow(key, data, type);
        }


    protected:

        template <class DataNavigatorT, class TypeNavigatorT>
        bool addValueRow(const std::string& key, DataNavigatorT& data, const TypeNavigatorT& type)
        {
            if (items.size() > 0)
            {
                QTreeWidgetItem* item = makeItem(key, data, type);
                items.top()->addChild(item);

                item->setData(columnKey, Qt::UserRole, serializePath(data.getPath()));
                item->setData(columnType, Qt::UserRole, int(type.getDescriptor()));

                if (false)
                {
                    QFont font;
                    font.setFamily("Consolas");
                    font.setStyleHint(QFont::Monospace);
                    font.setFixedPitch(true);
                    font.setPointSize(10);
                    item->setFont(columnValue, font);
                }
            }
            return true;
        }

        template <class DataNavigatorT, class TypeNavigatorT>
        QTreeWidgetItem* makeItem(const std::string& key, DataNavigatorT& data, const TypeNavigatorT& type) const
        {
            std::stringstream ss;
            try
            {
                this->streamValueText(data, type, ss);
            }
            catch (const aron::error::AronException& e)
            {
                ss << "x ";
                TreeDataVisitorBase::streamValueText(data, ss);
                std::stringstream es;
                es << e.what();
                ss << simox::alg::replace_all(es.str(), "\n", " | ");
            }
            return new QTreeWidgetItem(this->makeValueRowStrings(key, ss.str(), sanitizeTypeName(type.getName())));
        }

        QTreeWidgetItem* makeItem(const std::string& key, NDArrayDataNavigator& data, const PoseTypeNavigator& type) const;


        template <class DataNavigatorT, class TypeNavigatorT>
        void streamValueText(DataNavigatorT& data, const TypeNavigatorT& type, std::stringstream& ss) const
        {
            // Fallback to type-agnostic (but data-aware).
            (void) type;
            TreeDataVisitorBase::streamValueText(data, ss);
        }

        using TreeDataVisitorBase::streamValueText;
        void streamValueText(LongDataNavigator& data, const TimeTypeNavigator& type, std::stringstream& ss) const;
        void streamValueText(NDArrayDataNavigator& data, const PoseTypeNavigator& type, std::stringstream& ss) const;
        void streamValueText(NDArrayDataNavigator& data, const PositionTypeNavigator& type, std::stringstream& ss) const;
        void streamValueText(NDArrayDataNavigator& data, const OrientationTypeNavigator& type, std::stringstream& ss) const;


        std::string coeffSep = "  ";
    };

}

