#include "Segment.h"

#include <ArmarXCore/core/application/properties/PropertyDefinitionContainer.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/libraries/DebugObserverHelper/DebugObserverHelper.h>

#include <RobotAPI/libraries/aron/common/aron_conversions.h>
#include <RobotAPI/libraries/armem_robot_state/types.h>

#include <RobotAPI/libraries/armem/core/aron_conversions.h>
#include <RobotAPI/libraries/armem/core/MemoryID.h>
#include <RobotAPI/libraries/armem/client/Writer.h>
#include <RobotAPI/libraries/armem/client/query/Builder.h>
#include <RobotAPI/libraries/armem/client/query/query_fns.h>
#include <RobotAPI/libraries/armem/server/MemoryToIceAdapter.h>
#include <RobotAPI/libraries/armem/util/util.h>

#include <RobotAPI/libraries/armem_robot/robot_conversions.h>
#include <RobotAPI/libraries/armem_robot/aron/Robot.aron.generated.h>

#include <RobotAPI/libraries/armem_objects/types.h>
#include <RobotAPI/libraries/armem_objects/aron_conversions.h>
#include <RobotAPI/libraries/armem_robot_state/aron/Proprioception.aron.generated.h>


namespace armarx::armem::server::robot_state::proprioception
{

    Segment::Segment(armem::server::MemoryToIceAdapter& memoryToIceAdapter) :
        Base(memoryToIceAdapter, "Proprioception", arondto::Proprioception::toAronType(), 1024)
    {
    }

    Segment::~Segment() = default;


    void Segment::onConnect(RobotUnitInterfacePrx robotUnitPrx)
    {
        this->robotUnit = robotUnitPrx;

        std::string providerSegmentName = "Robot";
        KinematicUnitInterfacePrx kinematicUnit = robotUnit->getKinematicUnit();
        if (kinematicUnit)
        {
            providerSegmentName = kinematicUnit->getRobotName();
        }
        else
        {
            ARMARX_WARNING << "Robot unit '" << robotUnit->ice_getIdentity().name << "' does not have a kinematic unit."
                           << "\n Falling back to provider segment name '" << providerSegmentName << "'.";
        }
        this->robotUnitProviderID = segment->id().withProviderSegmentName(providerSegmentName);
    }



    Segment::RobotJointPositionMap Segment::getRobotJointPositionsLocking(
        const armem::Time& timestamp,
        DebugObserverHelper* debugObserver) const
    {
        return doLocked([this, &timestamp, &debugObserver]()
        {
            return getRobotJointPositions(timestamp, debugObserver);
        });
    }


    static
    aron::datanavigator::DictNavigatorPtr
    getDictElement(const aron::datanavigator::DictNavigator& dict, const std::string& key)
    {
        if (dict.hasElement(key))
        {
            return aron::datanavigator::DictNavigator::DynamicCastAndCheck(dict.getElement(key));
        }
        return nullptr;
    }


    Segment::RobotJointPositionMap Segment::getRobotJointPositions(
        const armem::Time& timestamp,
        DebugObserverHelper* debugObserver) const
    {
        namespace adn = aron::datanavigator;
        ARMARX_CHECK_NOT_NULL(segment);

        RobotJointPositionMap jointMap;

        TIMING_START(tProcessProviders);
        segment->forEachProviderSegment([&](const wm::ProviderSegment & provSeg)
        {
            TIMING_START(tProcessEntities);
            int i = 0;

            std::map<std::string, float>& robotJointMap = jointMap[provSeg.name()];
            provSeg.forEachEntity([&](const wm::Entity & entity)
            {
                TIMING_START(tProcessEntity);

                TIMING_START(tGetLatestInstance);
                if (entity.empty())
                {
                    return true;
                }
                const wm::EntitySnapshot& snapshot = entity.getLatestSnapshot();
                if (snapshot.empty())
                {
                    return true;
                }
                const wm::EntityInstance& entityInstance = snapshot.getInstance(0);
                TIMING_END_COMMENT_STREAM(tGetLatestInstance, "tGetLatestInstance " + std::to_string(i), ARMARX_DEBUG);

                TIMING_START(tCastAndEmplaceJointState);
                // Only cast the joint angles while avoiding throwing an exception.
                adn::DictNavigatorPtr joints = getDictElement(*entityInstance.data(), "joints");
                if (joints)
                {
                    adn::DictNavigatorPtr jointPositions = getDictElement(*joints, "position");
                    if (jointPositions)
                    {
                        for (const auto& [name, value] : jointPositions->getElements())
                        {
                            const float jointPosition = adn::FloatNavigator::DynamicCastAndCheck(value)->getValue();
                            robotJointMap.emplace(name, jointPosition);
                        }
                    }
                }
                TIMING_END_COMMENT_STREAM(tCastAndEmplaceJointState, "tCastAndEmplaceJointState " + std::to_string(i), ARMARX_DEBUG);

                TIMING_END_COMMENT_STREAM(tProcessEntity, "tProcessEntity " + std::to_string(i), ARMARX_DEBUG);

                if (debugObserver)
                {
                    debugObserver->setDebugObserverDatafield(dp + "t 1.1.1.1 GetLatestInstance (ms)", tGetLatestInstance.toMilliSecondsDouble());
                    debugObserver->setDebugObserverDatafield(dp + "t 1.1.1.2 CastAndEmplaceJointState (ms)", tCastAndEmplaceJointState.toMilliSecondsDouble());
                    debugObserver->setDebugObserverDatafield(dp + "t 1.1.1.T Process Entity (ms)", tProcessEntity.toMilliSecondsDouble());
                }
                ++i;

                return true;
            });

            TIMING_END_STREAM(tProcessEntities, ARMARX_DEBUG);
            if (debugObserver)
            {
                debugObserver->setDebugObserverDatafield(dp + "t 1.1 Process Entities (ms)", tProcessEntities.toMilliSecondsDouble());
            }
        });
        TIMING_END_STREAM(tProcessProviders, ARMARX_DEBUG);

        ARMARX_INFO << deactivateSpam(60) <<  "Number of known robot joint maps: " << jointMap.size();

        if (debugObserver)
        {
            debugObserver->setDebugObserverDatafield(dp + "t 1 Process Providers (ms)", tProcessProviders.toMilliSecondsDouble());
        }

        return jointMap;
    }


    const armem::MemoryID& Segment::getRobotUnitProviderID() const
    {
        return robotUnitProviderID;
    }

}  // namespace armarx::armem::server::robot_state::proprioception
