/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

// STD/STL
#include <memory>
#include <string>
#include <unordered_map>

// Base class
#include "../Serializer.h"

// ArmarX

namespace armarx::aron::cppserializer::detail
{
    template<typename TypenavigatorT, typename DerivedT>
    class SerializerBase :
        public cppserializer::Serializer
    {
    public:
        using PointerType = std::shared_ptr<DerivedT>;
        using TypenavigatorType = TypenavigatorT;

    public:
        SerializerBase(const std::string& cppName, const std::string& aronDataTypename, const std::string& aronTypeTypename, const typename TypenavigatorType::PointerType& t) :
            Serializer(cppName, aronDataTypename, aronTypeTypename),
            typenavigator(t)
        {
            ARMARX_CHECK_NOT_NULL(typenavigator);
        }

        virtual std::string getFullCppTypename() const override
        {
            switch (typenavigator->getMaybe())
            {
                case type::Maybe::eNone:
                    return getCoreCppTypename();
                case type::Maybe::eOptional:
                    return "std::optional<" + getCoreCppTypename() + ">";
                case type::Maybe::eRawPointer:
                    return getCoreCppTypename() + "*";
                case type::Maybe::eSharedPointer:
                    return "std::shared_ptr<" + getCoreCppTypename() + ">";
                case type::Maybe::eUniquePointer:
                    return "std::unique_ptr<" + getCoreCppTypename() + ">";
                default:
                    throw error::MaybeNotValidException("SerializerBase", "getFullCppTypename", "Received unknown maybe enum", typenavigator->getMaybe(), typenavigator->getPath());
            }
        }

        virtual std::string getFullTypenameGenerator() const override
        {
            switch (typenavigator->getMaybe())
            {
                case type::Maybe::eNone:
                    return getCoreCppTypename();
                case type::Maybe::eOptional:
                    return "std::make_optional<" + getCoreCppTypename() + ">";
                case type::Maybe::eRawPointer:
                    return "new " + getCoreCppTypename();
                case type::Maybe::eSharedPointer:
                    return "std::make_shared<" + getCoreCppTypename() + ">";
                case type::Maybe::eUniquePointer:
                    return "std::make_unique<" + getCoreCppTypename() + ">";
                default:
                    throw error::MaybeNotValidException("SerializerBase", "getFullTypenameGenerator", "Received unknown maybe enum", typenavigator->getMaybe(), typenavigator->getPath());
            }
        }

        virtual std::vector<CppFieldPtr> getPublicVariableDeclarations(const std::string& name) const override
        {
            CppFieldPtr field = CppFieldPtr(new CppField(this->getFullCppTypename(), name));
            return {field};
        }

        virtual std::vector<std::pair<std::string, std::string>> getCtorInitializers(const std::string&) const override
        {
            return {};
        }

        virtual CppBlockPtr getCtorBlock(const std::string&) const override
        {
            CppBlockPtr b = CppBlockPtr(new CppBlock());
            return b;
        }


    protected:
        std::string nextEl() const
        {
            switch (typenavigator->getMaybe())
            {
                case type::Maybe::eNone:
                    return ".";
                case type::Maybe::eOptional:
                case type::Maybe::eRawPointer:
                case type::Maybe::eSharedPointer:
                case type::Maybe::eUniquePointer:
                    return "->";
                default:
                    throw error::MaybeNotValidException("SerializerBase", "nextEl", "Received unknown maybe enum", typenavigator->getMaybe(), typenavigator->getPath());
            }
        }

        std::string toPointerAccessor(const std::string& accessor) const
        {
            switch (typenavigator->getMaybe())
            {
                case type::Maybe::eRawPointer:
                    return accessor;
                case type::Maybe::eSharedPointer:
                case type::Maybe::eUniquePointer:
                    return accessor + ".get()";
                default:
                    throw error::MaybeNotValidException("SerializerBase", "toPointerAccessor", "Received invalid maybe enum (not a pointer?)", typenavigator->getMaybe(), typenavigator->getPath());
            }
        }

    protected:
        typename TypenavigatorT::PointerType typenavigator;
    };
}
