/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "NavigatorReader.h"

// ArmarX
#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/core/navigator/data/AllNavigators.h>



namespace armarx::aron::dataIO::reader
{
    NavigatorReader::NavigatorReader(const data::AronDictPtr& n) :
        NavigatorReader(datanavigator::DictNavigator::DynamicCastAndCheck(datanavigator::Navigator::FromAronData(n)))
    {
    }

    NavigatorReader::NavigatorReader(const datanavigator::DictNavigatorPtr& n) :
        input(n)
    {
        auto current_nav_casted = datanavigator::DictNavigator::DynamicCastAndCheck(input);
        auto newToken = std::make_shared<NavigatorReaderToken>(current_nav_casted);
        stack.push(newToken);
    }

    // Containers
    ReaderInterface::ReadStartDictReturnType NavigatorReader::readStartDict()
    {
        if (!readFirstStartDict)
        {
            // we already added in constructor
            readFirstStartDict = true;
            auto lastToken = stack.top();
            auto current_nav = lastToken->getElement();
            int c = current_nav->childrenSize();
            return {c, true};
        }

        auto lastToken = stack.top();

        auto current_nav = lastToken->getNextElement();
        if (!current_nav)
        {
            lastToken->increaseCounter();
            return {0, false};
        }

        auto current_nav_casted = datanavigator::DictNavigator::DynamicCastAndCheck(current_nav);
        int c = current_nav->childrenSize();
        auto newToken = std::make_shared<NavigatorReaderToken>(current_nav_casted);
        stack.push(newToken);
        return {c, true};
    }

    bool NavigatorReader::readEndDict()
    {
        auto token = stack.top();
        datanavigator::DictNavigator::DynamicCastAndCheck(token->getElement());

        //std::cout << "We are at: " << token->getCurrentIndex() << "/" << token->getElementChildrenSize() << std::endl;

        if (token->finishedElement())
        {
            stack.pop();

            if (!stack.empty())
            {
                auto lastToken = stack.top();
                lastToken->increaseCounter();
            }

            return true;
        }
        return false;
    }

    ReaderInterface::ReadStartListReturnType NavigatorReader::readStartList()
    {
        auto lastToken = stack.top();

        auto current_nav = lastToken->getNextElement();
        if (!current_nav)
        {
            lastToken->increaseCounter();
            return {0, false};
        }

        auto current_nav_casted = datanavigator::ListNavigator::DynamicCastAndCheck(current_nav);
        int c = current_nav->childrenSize();
        auto newToken = std::make_shared<NavigatorReaderToken>(current_nav_casted);
        stack.push(newToken);
        return {c, true};
    }

    bool NavigatorReader::readEndList()
    {
        auto token = stack.top();
        datanavigator::ListNavigator::DynamicCastAndCheck(token->getElement());

        //std::cout << "We are at: " << token->getCurrentIndex() << "/" << token->getElementChildrenSize() << std::endl;
        if (token->finishedElement())
        {
            stack.pop();

            auto lastToken = stack.top();
            lastToken->increaseCounter();

            return true;
        }
        return false;
    }

    // Complex Data
    ReaderInterface::ReadStartNDArrayReturnType NavigatorReader::readStartNDArray()
    {
        auto lastToken = stack.top();

        auto current_nav = lastToken->getNextElement();
        if (!current_nav)
        {
            lastToken->increaseCounter();
            return {{}, "", false};
        }
        datanavigator::NDArrayNavigatorPtr casted = datanavigator::NDArrayNavigator::DynamicCastAndCheck(current_nav);
        return {casted->getDimensions(), casted->getType(), true};
    }

    void NavigatorReader::readEndNDArray(unsigned char* data)
    {
        if (data == NULL)
        {
            throw error::AronException("NavigatorReader", "readEndNDArray", "The corresponding data pointer of an complex aron ptr is NULL.");
        }

        auto lastToken = stack.top();
        auto current_nav = lastToken->getNextElement();
        lastToken->increaseCounter();

        auto casted = datanavigator::NDArrayNavigator::DynamicCastAndCheck(current_nav);
        std::vector<int> dims = casted->getDimensions();
        memcpy(data, casted->getData(), std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>()));
    }

    // Read primitives
    ReaderInterface::ReadPrimitiveReturnType NavigatorReader::readStartPrimitive()
    {
        auto lastToken = stack.top();
        auto current_nav = lastToken->getNextElement();
        if (!current_nav)
        {
            lastToken->increaseCounter();
            return {false};
        }
        return {true};
    }

    void NavigatorReader::readEndPrimitive(int& t)
    {
        auto lastToken = stack.top();
        auto current_nav = lastToken->getNextElement();
        lastToken->increaseCounter();

        auto casted = datanavigator::IntNavigator::DynamicCastAndCheck(current_nav);
        t = casted->getValue();
    }

    void NavigatorReader::readEndPrimitive(long& t)
    {
        auto lastToken = stack.top();
        auto current_nav = lastToken->getNextElement();
        lastToken->increaseCounter();

        auto casted = datanavigator::LongNavigator::DynamicCastAndCheck(current_nav);
        t = casted->getValue();
    }

    void NavigatorReader::readEndPrimitive(float& t)
    {
        auto lastToken = stack.top();
        auto current_nav = lastToken->getNextElement();
        lastToken->increaseCounter();

        auto casted = datanavigator::FloatNavigator::DynamicCastAndCheck(current_nav);
        t = casted->getValue();
    }

    void NavigatorReader::readEndPrimitive(double& t)
    {
        auto lastToken = stack.top();
        auto current_nav = lastToken->getNextElement();
        lastToken->increaseCounter();

        auto casted = datanavigator::DoubleNavigator::DynamicCastAndCheck(current_nav);
        t = casted->getValue();
    }

    void NavigatorReader::readEndPrimitive(std::string& t)
    {
        auto lastToken = stack.top();
        auto current_nav = lastToken->getNextElement();
        lastToken->increaseCounter();

        auto casted = datanavigator::StringNavigator::DynamicCastAndCheck(current_nav);
        t = casted->getValue();
    }

    void NavigatorReader::readEndPrimitive(bool& t)
    {
        auto lastToken = stack.top();
        auto current_nav = lastToken->getNextElement();
        lastToken->increaseCounter();

        auto casted = datanavigator::BoolNavigator::DynamicCastAndCheck(current_nav);
        t = casted->getValue();
    }

    std::string NavigatorReader::readKey()
    {
        auto token = stack.top();
        auto s = token->getCurrentKey();
        //std::cout << "Read key: " << s << std::endl;
        return s;
    }

    void NavigatorReader::loadMember(const std::string& k)
    {
        //std::cout << "Loading member " << k << std::endl;
        auto token = stack.top();
        token->setCurrentKey(k);
    }

    void NavigatorReader::readNull()
    {
        auto token = stack.top();
        auto current_nav = token->getNextElement();
        if (!current_nav)
        {
            throw error::AronException("NavigatorReader", "readNull", "The value of a navigator is not null!");
        }
        token->increaseCounter();
    }

    // Helper functions
    data::Descriptor NavigatorReader::getTypeOfNext() const
    {
        if (stack.empty())
        {
            return data::Descriptor::eDict;
        }
        auto token = stack.top();
        return token->getTypeOfNext();
    }
}
