/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "NlohmannJSONReader.h"

// ArmarX
#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/core/navigator/type/AllNavigators.h>

namespace armarx::aron::typeIO::reader
{
    NlohmannJSONReader::NlohmannJSONReader(const nlohmann::json& n) :
        input(n)
    {
    }

    NlohmannJSONReader::NlohmannJSONReader(const std::string& n) :
        input(nlohmann::json::parse(n))
    {
    }

    nlohmann::json NlohmannJSONReader::getNextAndIncrease()
    {
        if (!readInitialStart)
        {
            readInitialStart = true;
            return input;
        }
        NlohmannJSONReaderTokenPtr lastToken = stack.top();
        return lastToken->getNextElementAndIncreaseCnt();
    }

    ReaderInterface::ReadStartObjectReturnType NlohmannJSONReader::readStartObject()
    {
        nlohmann::json current_json = getNextAndIncrease();
        auto newToken = std::make_shared<NlohmannJSONReaderToken>(type::Descriptor::eObject, current_json);
        stack.push(newToken);
        type::Maybe o = current_json[io::Data::READER_WRITER_MAYBE_SLUG];
        return {newToken->getElementName(), newToken->getElementChildrenSize(), o};
    }

    bool NlohmannJSONReader::readEndObject()
    {
        NlohmannJSONReaderTokenPtr token = stack.top();
        token->assertType(type::Descriptor::eObject);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    ReaderInterface::ReadStartDictReturnType NlohmannJSONReader::readStartDict()
    {
        nlohmann::json current_json = getNextAndIncrease();
        auto newToken = std::make_shared<NlohmannJSONReaderToken>(type::Descriptor::eDict, current_json);
        stack.push(newToken);
        type::Maybe o = current_json[io::Data::READER_WRITER_MAYBE_SLUG];
        return {newToken->getElementChildrenSize(), o};
    }

    bool NlohmannJSONReader::readEndDict()
    {
        NlohmannJSONReaderTokenPtr token = stack.top();
        token->assertType(type::Descriptor::eDict);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    ReaderInterface::ReadStartListReturnType NlohmannJSONReader::readStartList()
    {
        nlohmann::json current_json = getNextAndIncrease();
        auto newToken = std::make_shared<NlohmannJSONReaderToken>(type::Descriptor::eList, current_json);
        stack.push(newToken);
        type::Maybe o = current_json[io::Data::READER_WRITER_MAYBE_SLUG];
        return {newToken->getElementChildrenSize(), o};
    }

    bool NlohmannJSONReader::readEndList()
    {
        NlohmannJSONReaderTokenPtr token = stack.top();
        token->assertType(type::Descriptor::eList);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    ReaderInterface::ReadStartTupleReturnType NlohmannJSONReader::readStartTuple()
    {
        nlohmann::json current_json = getNextAndIncrease();
        auto newToken = std::make_shared<NlohmannJSONReaderToken>(type::Descriptor::eTuple, current_json);
        stack.push(newToken);
        type::Maybe o = current_json[io::Data::READER_WRITER_MAYBE_SLUG];
        return {newToken->getElementChildrenSize(), o};
    }

    bool NlohmannJSONReader::readEndTuple()
    {
        NlohmannJSONReaderTokenPtr token = stack.top();
        token->assertType(type::Descriptor::eTuple);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    ReaderInterface::ReadStartPairReturnType NlohmannJSONReader::readStartPair()
    {
        nlohmann::json current_json = getNextAndIncrease();
        auto newToken = std::make_shared<NlohmannJSONReaderToken>(type::Descriptor::ePair, current_json);
        stack.push(newToken);
        type::Maybe o = current_json[io::Data::READER_WRITER_MAYBE_SLUG];
        return {newToken->getElementChildrenSize(), o};
    }

    bool NlohmannJSONReader::readEndPair()
    {
        NlohmannJSONReaderTokenPtr token = stack.top();
        token->assertType(type::Descriptor::ePair);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    ReaderInterface::ReadEigenMatrixReturnType NlohmannJSONReader::readEigenMatrix()
    {
        nlohmann::json j = getNextAndIncrease();
        std::vector<unsigned int> dims = j[io::Data::READER_WRITER_NDARRAY_DIMENSIONS_SLUG];
        std::string type = j[io::Data::READER_WRITER_NDARRAY_TYPE_SLUG];
        type::Maybe o = j[io::Data::READER_WRITER_MAYBE_SLUG];
        return {dims[0], dims[1], type, o};
    }

    ReaderInterface::ReadEigenQuaternionReturnType NlohmannJSONReader::readEigenQuaternion()
    {
        nlohmann::json j = getNextAndIncrease();
        std::string type = j[io::Data::READER_WRITER_NDARRAY_TYPE_SLUG];
        type::Maybe o = j[io::Data::READER_WRITER_MAYBE_SLUG];
        return {type, o};
    }

    ReaderInterface::ReadIVTCbyteImageReturnType NlohmannJSONReader::readIVTCByteImage()
    {
        nlohmann::json j = getNextAndIncrease();
        type::Maybe o = j[io::Data::READER_WRITER_MAYBE_SLUG];
        return {o};
    }

    ReaderInterface::ReadOpenCVMatReturnType NlohmannJSONReader::readOpenCVMat()
    {
        nlohmann::json j = getNextAndIncrease();
        type::Maybe o = j[io::Data::READER_WRITER_MAYBE_SLUG];
        return {o};
    }

    ReaderInterface::ReadPCLPointCloudReturnType NlohmannJSONReader::readPCLPointCloud()
    {
        nlohmann::json j = getNextAndIncrease();
        std::string type = j[io::Data::READER_WRITER_NDARRAY_TYPE_SLUG];
        type::Maybe o = j[io::Data::READER_WRITER_MAYBE_SLUG];
        return {type, o};
    }

    ReaderInterface::ReadPositionReturnType NlohmannJSONReader::readPosition()
    {
        nlohmann::json j = getNextAndIncrease();
        type::Maybe o = j[io::Data::READER_WRITER_MAYBE_SLUG];
        return {o};
    }

    ReaderInterface::ReadOrientationReturnType NlohmannJSONReader::readOrientation()
    {
        nlohmann::json j = getNextAndIncrease();
        type::Maybe o = j[io::Data::READER_WRITER_MAYBE_SLUG];
        return {o};
    }

    ReaderInterface::ReadPoseReturnType NlohmannJSONReader::readPose()
    {
        nlohmann::json j = getNextAndIncrease();
        type::Maybe o = j[io::Data::READER_WRITER_MAYBE_SLUG];
        return {o};
    }

    // Read primitives
    ReaderInterface::ReadPrimitiveReturnType NlohmannJSONReader::readInt()
    {
        nlohmann::json j = getNextAndIncrease();
        type::Maybe o = j[io::Data::READER_WRITER_MAYBE_SLUG];
        return {o};
    }

    ReaderInterface::ReadPrimitiveReturnType NlohmannJSONReader::readLong()
    {
        nlohmann::json j = getNextAndIncrease();
        type::Maybe o = j[io::Data::READER_WRITER_MAYBE_SLUG];
        return {o};
    }

    ReaderInterface::ReadPrimitiveReturnType NlohmannJSONReader::readFloat()
    {
        nlohmann::json j = getNextAndIncrease();
        type::Maybe o = j[io::Data::READER_WRITER_MAYBE_SLUG];
        return {o};
    }

    ReaderInterface::ReadPrimitiveReturnType NlohmannJSONReader::readDouble()
    {
        nlohmann::json j = getNextAndIncrease();
        type::Maybe o = j[io::Data::READER_WRITER_MAYBE_SLUG];
        return {o};
    }

    ReaderInterface::ReadPrimitiveReturnType NlohmannJSONReader::readString()
    {
        nlohmann::json j = getNextAndIncrease();
        type::Maybe o = j[io::Data::READER_WRITER_MAYBE_SLUG];
        return {o};
    }

    ReaderInterface::ReadPrimitiveReturnType NlohmannJSONReader::readBool()
    {
        nlohmann::json j = getNextAndIncrease();
        type::Maybe o = j[io::Data::READER_WRITER_MAYBE_SLUG];
        return {o};
    }

    ReaderInterface::ReadPrimitiveReturnType NlohmannJSONReader::readTime()
    {
        nlohmann::json j = getNextAndIncrease();
        type::Maybe o = j[io::Data::READER_WRITER_MAYBE_SLUG];
        return {o};
    }

    std::string NlohmannJSONReader::readKey()
    {
        auto token = stack.top();
        return token->getCurrentKey();
    }

    void NlohmannJSONReader::loadMember(const std::string& k)
    {
        auto token = stack.top();
        token->setCurrentKey(k);
    }

    // Helper functions
    type::Descriptor NlohmannJSONReader::getTypeOfNext() const
    {
        if (stack.empty())
        {
            return type::Descriptor::eObject;
        }

        auto token = stack.top();
        return token->getTypeOfNext();
    }
}
